"""
    IQtools - Utilities for IQ data visualization/manipulation in Python
    Copyright (C) 2023  Logan Fagg

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

# Source file for IQtools functions

from enum import Enum
import numpy as np
from .IQdata import IQdata


class DiscreteSignal():
    def __init__(self, description, params: {}):
        self.description = description
        self.params = params

    def __str__(self):
        info = self.description + ' ('
        for p,v in self.params.items():
            info+=p+':'+str(v)+', '
        info = info[:-2]+')'
        return info


class SignalGenerator():

    def __init__(self, sampleRate, bits, duration):
        self.sampleRate = sampleRate
        self.bits = bits
        self.duration = duration
        self.t = np.arange(0, self.duration, 1/self.sampleRate)
        self.s = np.zeros(np.size(self.t))
        self.FS = (2**(self.bits-1))-1

    def addSinusoid(self,amplitude_Rel_FS, frequency, phaseRads = 0):
        newSig = amplitude_Rel_FS*self.FS*(np.cos(frequency*2*3.14159*self.t+phaseRads)+1j*np.sin(frequency*2*3.14159*self.t+phaseRads))
        self.s = self.s + newSig

    def saveToFile(self, filename):
        with open(filename, 'w') as outPutFile:
            for sample in self.s:
                idata = sample.real
                qdata = sample.imag
                outPutFile.write(str(int(idata))+','+str(int(qdata))+'\n')



"""
    IQtools - Utilities for IQ data visualization/manipulation in Python
    Copyright (C) 2023  Logan Fagg

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from enum import Enum


class DataFormats(Enum):
    IQ = 0
    INTERP = 1
    SEQ = 2


def load_IQ(file, bits):
    samples = []
    scale = (2 ** (bits - 1))
    for line in file:
        splitLine = line.split(',')
        value = (complex(float(splitLine[0].strip()) / scale, float(splitLine[1].strip()) / scale))
        samples.append(value)
    return samples


class IQdata:
    def __init__(self, fileName: str, sampleRate: float, bits=0, format=DataFormats.IQ, refOffset=0):
        """
        Initializes a data structure for IQ samples and their metadata
        :param fileName: Source datafile TODO: Add support for nonfile input
        :param sampleRate: Complex samplerate of the data
        :param bits:  Bitness of the data (assumes signed data)
        :param format: Not currently used
        :param refOffset: Not currently used
        """
        self.sampleRate = sampleRate
        self.bits = bits
        self.format = format
        self.refOffset = refOffset
        self.loadData(fileName)
        self.datalen = len(self.samples)

    def loadData(self, fileName: str) -> None:
        """
        Helper function for loading samples from a file
        :param fileName:
        """
        file = open(fileName, 'r')
        if self.format == DataFormats.IQ:
            self.samples = load_IQ(file, self.bits)
        else:
            raise ValueError('Unsupported IQ File Format Supplied')


"""
    IQtools - Utilities for IQ data visualization/manipulation in Python
    Copyright (C) 2023  Logan Fagg

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

# Source file for IQtools functions

from enum import Enum
import numpy as np
from scipy.signal import get_window
import math
from .IQdata import IQdata

class SignalAnalyzer():
    def __init__(self, data: IQdata, ref_level=0, ref_unit='dBFS'):
        """
        Class requires an IQdata object containing samples and their metadata
        :param data:
        """
        self.window = 'boxcar'
        self.ref_level = ref_level
        self.ref_unit = ref_unit
        self.dataT = data
        self.t = np.arange(0, self.dataT.datalen * (1 / self.dataT.sampleRate), 1 / self.dataT.sampleRate)
        self.getDataF()

    def getDataF(self, windowName='boxcar'):
        self.window = windowName
        window = get_window(windowName, self.dataT.datalen)
        windowedData = self.dataT.samples * window
        self.dataF = np.fft.fft(windowedData) / self.dataT.datalen
        self.f = np.fft.fftfreq(self.dataT.datalen) * (self.dataT.sampleRate)

    def getSpectrumMag(self, frequencyBase_Hz = 1):
        """
        Process spectral data into dBFS magnitude
        TODO: add support for basic manipulation of data scaling/offset
        :rtype: tuple of frequency array and array of dBFS spectrum values
        """
        magData = 20 * np.log10(np.abs(self.dataF)) + self.ref_level
        freqData = self.f/frequencyBase_Hz
        return freqData, magData

    def getPower(self, maxpower=np.inf):
        maglin = np.abs(self.dataF)**2
        maxpowerLin = 10**(maxpower/10)
        maglinFiltered = []
        for val in maglin:
            if(val>maxpowerLin):
                maglinFiltered+=[0]
            else:
                maglinFiltered+=[val]

        power = np.sqrt(sum(maglinFiltered))
        window = get_window(self.window, self.dataT.datalen)

        ecf = 1/np.sqrt(sum(window**2)/self.dataT.datalen)

        power = power*ecf

        return 20*np.log10(power)

    def getSubSpectrumMag(self, left, right, exclusiveLeft = True, exclusiveRight = False):
        # TODO: Implement
        pass


    def getPeakList(self, minpower = -np.inf, frequencyBase_Hz=1):
        f, magData = self.getSpectrumMag(frequencyBase_Hz=frequencyBase_Hz)
        prevSample = minpower
        samples = len(magData)
        peaks = []
        peaks_f = []
        for index in range(samples):
            sample = magData[index]
            if (sample < prevSample) and (prevSample > minpower):
                peaks.append(prevSample)
                peaks_f.append(self.f[index-1]/frequencyBase_Hz)
            prevSample = sample
        return peaks, peaks_f


import argparse

from .IQdata import IQdata
from .SignalAnalyzer import SignalAnalyzer
from .SignalGenerator import SignalGenerator
import matplotlib.pyplot as plt




def getPeaks(args=None):
    parser = argparse.ArgumentParser(description='Utility for getting peak spectrum values from an IQ data file')
    parser.add_argument('filename', help='File with IQ data', type=str)
    parser.add_argument('sampleRate', help='Complex sample rate of the file', type=float)
    parser.add_argument('-f', help='Frequency Units', type=float, default=1E6)
    parser.add_argument('-b', help='Number of signed bits for the ADC', type=int, default=12)
    parser.add_argument('-t', help='Threshold for peak detection (In dBFS by default)', type=float, default=-float('inf'))

    config = parser.parse_args(args)

    bits = config.b
    sampleRate = config.sampleRate
    filename = config.filename
    threshold = config.t
    frequencyUnits = config.f
    data = IQdata(filename, sampleRate, bits)
    sa = SignalAnalyzer(data)
    totalPower = sa.getPower()
    peaks, peaksF = sa.getPeakList(minpower=threshold, frequencyBase_Hz=frequencyUnits)
    print('Total Power: {0:.2f} dBFS'.format(totalPower))
    peaksFound = len(peaks)
    for index, peak, frequency in zip(range(peaksFound), peaks, peaksF):
        print('{index}: {peak:.2f} dBFS @ {frequency: 0.2f} MHz'.format(index=index, peak=peak, frequency=frequency))

def getSpectrum(args=None):
    parser = argparse.ArgumentParser(description='Utility for generating a frequency spectrum from an IQ data file')
    parser.add_argument('filename', help='File with IQ data', type=str)
    parser.add_argument('sampleRate', help='Complex sample rate of the file', type=float)
    parser.add_argument('-f', help='Frequency Units', type=float, default=1E6)
    parser.add_argument('-b', help='Number of signed bits for the ADC', type=int, default=12)
    parser.add_argument('-s', help='Save the image to a png instead of showing it', action='store_true')


    config = parser.parse_args(args)
    bits = config.b
    sampleRate = config.sampleRate
    filename = config.filename
    frequencyUnits = config.f
    data = IQdata(filename, sampleRate, bits)
    saveFig = config.s
    sa = SignalAnalyzer(data)
    f, mag = sa.getSpectrumMag()
    plt.plot(f, mag)
    plt.ylim((-100, 0))
    plt.ylabel('dBFS')
    plt.xlabel('MHz')
    plt.title(filename.split('.')[0])
    plt.tight_layout()
    if saveFig:
        plt.savefig(filename.split('.')[0]+'.png')
    else:
        plt.show()
        plt.clf()

def generateCW(args=None):
    parser = argparse.ArgumentParser(description='Utility for generating a CW IQ signal')
    parser.add_argument('filename', help='Output File', type=str)
    parser.add_argument('level', help='Output amplitude (0 to 1.0)', type=float)
    parser.add_argument('baseBandFrequency', help='Baseband frequency (must be within Nyquist, +/- Fs/2)', type=float)
    parser.add_argument('sampleRate', help='Complex sample rate of the file', type=float)
    parser.add_argument('duration', help='Duration of sample file in milliseconds', type=float)
    parser.add_argument('-p', help='phase offset (rads)', type=float, default=0)
    parser.add_argument('-b', help='Number of signed bits for the ADC', type=int, default=12)



    config = parser.parse_args(args)
    bits = config.b
    sampleRate = config.sampleRate
    filename = config.filename
    level = config.level
    duration = config.duration
    phaseOffset = config.p
    bbFreq = config.baseBandFrequency
    sg = SignalGenerator(sampleRate, bits, duration/1000)
    sg.addSinusoid(level, bbFreq, phaseRads=phaseOffset)
    sg.saveToFile(filename)

